/*
 * ghttp.c
 *
 * Copyright (C) 1993-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "port_before.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "port_after.h"

#include "mime.h"

struct HTTPContextP
{
  MemPool mp, rmp;
  MIMEContext mc;
  bool checked_header;  /* indicates header lines parsed */
  bool got_magic;       /* indicates magic line parsed */

  size_t hoff;          /* offset to header */
  size_t doff;          /* offset to data */

  byte *b;              /* buffer being parsed */
  size_t blen;          /* length of buffer */
};

static void CheckData _ArgProto((HTTPInfo *));

/*
 * HTTPCheckHeader
 *
 * Scrounges around in the header fields to see if there is any
 * interesting information.
 */
static int
HTTPCheckHeader(hi)
HTTPInfo *hi;
{
  bool cache;
  char *value;
  size_t clen;
  char *option;
  char *list;
  URLParts *up;
  char *url;
  int status;

  status = MIMEGetHTTPStatus(hi->mc);
  if (status >= 300) cache = false;
  else cache = true;

  /*
   * Take action on MIME fields.
   */
  if ((value = MIMEGetField(hi->mc, "location")) != NULL &&
      status >= 300 && status < 400)
  {
    up = URLParse(hi->mp, value);
    
    /* Just blindly follow the URL unless it should be used as a proxy */
    if (status != 305)
    {
      hi->url = MPStrDup(hi->mp, value);
      hi->up = up;
    }
    else hi->pup = up;
    
    if (HTTPOpen(hi) == -1) HTTPFailure(hi);
    
    return(-1);
  }
  if ((value = MIMEGetField(hi->mc, "content-length")) != NULL)
  {
    clen = (size_t)atoi(value) + hi->doff;
    if (clen > 0 && clen > hi->blen && clen > hi->bsize)
    {
      hi->b = (byte *)realloc_mem(hi->b, clen);
      hi->bsize = clen;
      hi->bmax = clen;
    }
    else hi->bmax = 0;
  }
  if ((value = MIMEGetField(hi->mc, "pragma")) != NULL)
  {
    list = value;
    while ((option = mystrtok(list, ' ', &list)) != NULL)
    {
      if (strcasecmp(option, "no-cache") == 0) cache = false;
    }
  }
  if ((value = MIMEGetField(hi->mc, "www-authenticate")) != NULL)
  {
    cache = false;
  }
  
  if (hi->url != NULL) url = hi->url;
  else if (hi->wr->url != NULL) url = hi->wr->url;
  else url = "unknown";
  
  /* Do not call this before dealing with "Location:" */
/*
  SourceInit(hi->ws, cache);
*/
  
  return(0);
}

/*
 * CheckData
 */
static void
CheckData(hi)
HTTPContext hi;
{
  char *s;
  size_t i;

  if (!hi->got_magic)
  {
    s = (char *)hi->b;
    if (s[0] != 'H' || s[1] != 'T' || s[2] != 'T' || s[3] != 'P')
    {
      hi->got_magic = true;
      hi->got_header = true;
    }
    else
    {
      for (; i < blen; i++, s++)
      {
	if (*s == '\n')
	{
	  hi->hoff = i + 1;
	  hi->got_magic = true;
	  hi->mc = MIMECreateContext();
	  break;
	}
      }
    }
  }
  else if (!hi->got_header)
  {
    MIMEAddData(hi->mc, hi->b + hi->hoff, hi->blen - hi->hoff);
  }

  return;
}

/*
 * HTTPAddData
 */
void
HTTPAddData(hi, b, blen)
HTTPContext hi;
byte *b;
size_t blen;
{
  myassert(hi->blen <= blen, "hi->blen > blen!");

  if (blen == 0) return;

  hi->b = b;
  hi->blen = blen;

  CheckData(hi);

  return;
}

/*
 * HTTPEndData
 */
void
HTTPEndData(hi, b, blen)
HTTPContext hi;
byte *b;
size_t blen;
{
  myassert(hi->blen <= blen, "hi->blen > blen!");

  if (blen == 0) return;

  hi->b = b;
  hi->blen = blen;

  CheckData(hi);

  return;
}

/*
 * HTTPDestroyContext
 */
void
HTTPDestroyContext(hi)
HTTPContext hi;
{
  MPDestroy(hi->rmp);
  MPDestroy(hi->mp);
  return;
}

/*
 * HTTPCreateContext
 */
HTTPContext
HTTPCreateContext(dataproc, headerproc, errorproc, closure)
HTTPDataProc dataproc;
HTTPHeaderProc headerproc;
HTTPErrorProc errorproc;
void *closure;
{
  HTTPContext hi;
  MemPool mp;

  mp = MPCreate();
  hi = (HTTPContext)MPCGet(mp, sizeof(struct HTTPContextP));
  hi->mp = mp;
  hi->rmp = MPCreate();

  return(hi);
}
